"""Manager for pyhf integration"""
from typing import Text, List
import logging, importlib

from spey.system.exceptions import MethodNotAvailable


class PyhfManager:
    """Manager to unify the usage of pyhf through out the package"""

    pyhf = importlib.import_module("pyhf")

    def __init__(self):
        PyhfManager.pyhf.set_backend("numpy", precision="64b")

        PyhfManager.pyhf.pdf.log.setLevel(logging.CRITICAL)
        PyhfManager.pyhf.workspace.log.setLevel(logging.CRITICAL)

        self.shim = importlib.import_module("pyhf.optimize.common").shim

    def __repr__(self) -> Text:
        return (
            f"pyhfManager(pyhf_backend='{self.backend}',"
            + f" available_backends={self.available_backends})"
        )

    @property
    def grad_available(self) -> bool:
        """
        Returns true or false depending on the availablility
        of the gradient within the current backend.
        """
        return self.backend != "numpy"

    @property
    def available_backends(self) -> List[Text]:
        """
        Retreive the names of available backends

        Returns:
            ``List[Text]``:
            list of available backends
        """
        return [
            bd for bd in ["numpy", "tensorflow", "jax"] if importlib.util.find_spec(bd)
        ]

    @property
    def backend(self) -> Text:
        """Retreive current backend name"""
        return PyhfManager.pyhf.tensorlib.name

    @backend.setter
    def backend(self, backend: Text) -> None:
        """
        Modify pyhf backend.

        Args:
            backend (`Text`): backend type. see pyhf for details
        """
        if backend not in self.available_backends:
            raise MethodNotAvailable(f"{backend} backend currently not available.")
        PyhfManager.pyhf.set_backend(backend)

    @property
    def backend_accessor(self):
        """access to current pyhf backend"""
        return importlib.import_module(self.backend)



from typing import Optional, List, Tuple, Dict, Text, Union

from dataclasses import dataclass
from abc import ABC, abstractmethod
import json, copy
import numpy as np

from spey.base import ModelConfig
from spey import ExpectationType

from . import manager


class Base(ABC):
    """Base class for pyhf data input"""

    def __call__(self, expected: ExpectationType = ExpectationType.observed) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        return None, None, None

    @abstractmethod
    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """

    @property
    @abstractmethod
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""


@dataclass
class SimpleModelData(Base):
    """
    Dataclass for simple statistical model

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[float]``): observations
        absolute_uncertainties (``List[float]``): absolute uncertainties on the background
    """

    signal_yields: List[float]
    background_yields: List[float]
    data: List[float]
    absolute_uncertainties: List[float]

    def __post_init__(self):
        assert (
            len(self.signal_yields)
            == len(self.background_yields)
            == len(self.data)
            == len(self.absolute_uncertainties)
        ), "Lenght of all input must be the same."

    def __call__(self, expected: ExpectationType = ExpectationType.observed) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        model = manager.pyhf.simplemodels.uncorrelated_background(
            self.signal_yields, self.background_yields, self.absolute_uncertainties
        )

        data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        ) + model.config.auxdata

        return None, model, data

    @property
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return any(x > 0 for x in self.signal_yields)

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        signal = np.array(self.signal_yields)
        nb = np.array(self.background_yields)
        minimum_poi = (
            -np.min(np.true_divide(nb[signal != 0.0], signal[signal != 0.0])).astype(
                np.float32
            )
            if len(signal[signal != 0.0]) > 0
            else -np.inf
        )

        _, model, _ = self()

        bounds = model.config.suggested_bounds()
        bounds[model.config.poi_index] = (
            max(minimum_poi, -10.0) if allow_negative_signal else 0.0,
            bounds[model.config.poi_index][1] if not poi_upper_bound else poi_upper_bound,
        )

        return ModelConfig(
            poi_index=model.config.poi_index,
            minimum_poi=minimum_poi,
            suggested_init=model.config.suggested_init(),
            suggested_bounds=bounds,
            parameter_names=model.config.par_names,
        )


@dataclass
class FullStatisticalModelData(Base):
    """
    Data container for the full statistical model.

    Args:
        signal_patch (``List[Dict]``): Patch data for signal model. please see
            `this link <https://pyhf.readthedocs.io/en/v0.7.0/likelihood.html>`_ for details on
            the structure of the input.
        background_only_model (``Dict`` or ``Text``): This input expects background only data
            that describes the full statistical model for the background. It also accepts ``str``
            input which indicates the full path to the background only ``JSON`` file.
    """

    signal_patch: List[Dict]
    background_only_model: Union[Dict, Text]

    def __post_init__(self):
        if isinstance(self.background_only_model, str):
            with open(self.background_only_model, "r", encoding="uft-8") as f:
                self.background_only_model = json.load(f)

        self.background_only_model_apriori = copy.deepcopy(self.background_only_model)

        # set data as expected background events
        obs = []
        for channel in self.background_only_model_apriori.get("channels", []):
            current = []
            for ch in channel["samples"]:
                if len(current) == 0:
                    current = [0.0] * len(ch["data"])
                current = [cur + dt for cur, dt in zip(current, ch["data"])]
            obs.append({"name": channel["name"], "data": current})
        self.background_only_model_apriori["observations"] = obs

        self.workspace_apriori = manager.pyhf.Workspace(
            self.background_only_model_apriori
        )
        self.workspace = manager.pyhf.Workspace(self.background_only_model)

        min_ratio = []
        for idc, channel in enumerate(self.background_only_model.get("channels", [])):
            current_signal = []
            for sigch in self.signal_patch:
                if idc == int(sigch["path"].split("/")[2]):
                    current_signal = np.array(
                        sigch.get("value", {}).get("data", []), dtype=np.float32
                    )
                    break
            if len(current_signal) == 0:
                continue
            current_bkg = []
            for ch in channel["samples"]:
                if len(current_bkg) == 0:
                    current_bkg = np.zeros(shape=(len(ch["data"]),), dtype=np.float32)
                current_bkg += np.array(ch["data"], dtype=np.float32)
            min_ratio.append(
                np.min(
                    np.true_divide(
                        current_bkg[current_signal != 0.0],
                        current_signal[current_signal != 0.0],
                    )
                )
                if np.any(current_signal != 0.0)
                else np.inf
            )
        self._minimum_poi = (
            -np.min(min_ratio).astype(np.float32) if len(min_ratio) > 0 else -np.inf
        )

    def __call__(self, expected: ExpectationType = ExpectationType.observed) -> Tuple:
        """
        Retreive pyhf workspace

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Returns:
            ``Tuple``:
            workspace, model and data
        """
        if expected == ExpectationType.apriori:
            model = self.workspace_apriori.model(
                patches=[self.signal_patch],
                modifier_settings={
                    "normsys": {"interpcode": "code4"},
                    "histosys": {"interpcode": "code4p"},
                },
            )

            return self.workspace_apriori, model, self.workspace_apriori.data(model)

        model = self.workspace.model(
            patches=[self.signal_patch],
            modifier_settings={
                "normsys": {"interpcode": "code4"},
                "histosys": {"interpcode": "code4p"},
            },
        )
        return self.workspace, model, self.workspace.data(model)

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        _, model, _ = self()

        bounds = model.config.suggested_bounds()
        bounds[model.config.poi_index] = (
            max(self._minimum_poi, -10.0) if allow_negative_signal else 0.0,
            bounds[model.config.poi_index][1] if not poi_upper_bound else poi_upper_bound,
        )

        return ModelConfig(
            poi_index=model.config.poi_index,
            minimum_poi=self._minimum_poi,
            suggested_init=model.config.suggested_init(),
            suggested_bounds=bounds,
            parameter_names=model.config.par_names,
        )

    @property
    def isAlive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        for channel in (
            ch for ch in self.signal_patch if ch.get("value", None) is not None
        ):
            if any(nsig > 0.0 for nsig in channel["value"].get("data", [])):
                return True
        return False



from typing import Dict, Union, Optional, Tuple, List, Any, Callable

import warnings
import numpy as np

from spey.utils import ExpectationType
from spey.system.exceptions import InvalidInput

from . import manager

__all__ = ["initialise_workspace"]


def initialise_workspace(
    signal: Union[List[float], List[Dict]],
    background: Union[Dict, List[float]],
    nb: Optional[List[float]] = None,
    delta_nb: Optional[List[float]] = None,
    expected: ExpectationType = ExpectationType.observed,
    return_full_data: bool = False,
) -> Union[
    tuple[
        Union[list, Any],
        Union[Optional[dict], Any],
        Optional[Any],
        Optional[Any],
        Optional[manager.pyhf.Workspace],
        Any,
        Any,
        Union[Union[int, float, complex], Any],
    ],
    tuple[Optional[manager.pyhf.Workspace], Any, Any],
]:
    """
    Construct the statistical model with respect to the given inputs.

    Args:
        signal (``Union[List[float], List[Dict]]``): signal yields
        background (``Union[Dict, List[float]]``): background yields or background only json file
        nb (``Optional[List[float]]``, default ``None``): number of background yields. This input
          is only used if ``signal`` and ``background`` input are ``List[float]``.
        delta_nb (``Optional[List[float]]``, default ``None``): uncertainty on background.This input
          is only used if ``signal`` and ``background`` input are ``List[float]``.
        expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
            p-values to be computed.

            * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
            prescriotion which means that the experimental data will be assumed to be the truth
            (default).
            * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
            post-fit prescriotion which means that the experimental data will be assumed to be
            the truth.
            * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
            prescription which means that the SM will be assumed to be the truth.
        return_full_data (``bool``, default ``False``): _description_

    Raises:
        ``InvalidInput``: _description_

    Returns:
        ``Union[ tuple[ Union[list, Any], Union[Optional[dict], Any], Optional[Any], Optional[Any], Optional[manager.pyhf.Workspace], Any, Any, Union[Union[int, float, complex], Any], ], tuple[Optional[manager.pyhf.Workspace], Any, Any], ]``:
        _description_
    """
    # Check the origin of signal
    signal_from_patch = False
    if isinstance(signal, (float, np.ndarray)):
        signal = np.array(signal, dtype=np.float32).reshape(-1)
    elif isinstance(signal, list):
        if isinstance(signal[0], dict):
            signal_from_patch = True
        else:
            signal = np.array(signal, dtype=np.float32).reshape(-1)
    else:
        raise InvalidInput(f"An unexpected type of signal has been presented: {signal}")

    # check the origin of background
    bkg_from_json = False
    if isinstance(background, dict):
        bkg_from_json = True
    elif isinstance(background, (float, list, np.ndarray)):
        background = np.array(background, dtype=np.float32).reshape(-1)
    else:
        raise InvalidInput(
            f"An unexpected type of background has been presented: {background}"
        )

    if (bkg_from_json and not signal_from_patch) or (
        signal_from_patch and not bkg_from_json
    ):
        raise InvalidInput("Signal and background types does not match.")

    if not bkg_from_json:
        # check if bkg uncertainties are valid
        if isinstance(delta_nb, (float, list, np.ndarray)):
            delta_nb = np.array(delta_nb, dtype=np.float32).reshape(-1)
        else:
            raise InvalidInput(
                f"An unexpected type of background uncertainty has been presented: {delta_nb}"
            )
        # check if MC bkg data is valid
        if isinstance(nb, (float, list, np.ndarray)):
            nb = np.array(nb, dtype=np.float32).reshape(-1)
        else:
            raise InvalidInput(
                f"An unexpected type of background has been presented: {nb}"
            )
        assert (
            len(signal) == len(background) == len(nb) == len(delta_nb)
        ), "Dimensionality of the data does not match."
    else:
        delta_nb, nb = None, None

    workspace, model, data, minimum_poi = None, None, None, -np.inf

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        if not bkg_from_json:
            if expected == ExpectationType.apriori:
                # set data as expected background events
                background = nb
            # Create model from uncorrelated region
            model = manager.pyhf.simplemodels.uncorrelated_background(
                signal.tolist(), nb.tolist(), delta_nb.tolist()
            )
            data = background.tolist() + model.config.auxdata

            if return_full_data:
                minimum_poi = (
                    -np.min(
                        np.true_divide(nb[signal != 0.0], signal[signal != 0.0])
                    ).astype(np.float32)
                    if len(signal[signal != 0.0]) > 0
                    else -np.inf
                )

        else:
            if expected == ExpectationType.apriori:
                # set data as expected background events
                obs = []
                for channel in background.get("channels", []):
                    current = []
                    for ch in channel["samples"]:
                        if len(current) == 0:
                            current = [0.0] * len(ch["data"])
                        current = [cur + dt for cur, dt in zip(current, ch["data"])]
                    obs.append({"name": channel["name"], "data": current})
                background["observations"] = obs

            workspace = manager.pyhf.Workspace(background)
            model = workspace.model(
                patches=[signal],
                modifier_settings={
                    "normsys": {"interpcode": "code4"},
                    "histosys": {"interpcode": "code4p"},
                },
            )

            data = workspace.data(model)

            if return_full_data and None not in [model, workspace, data]:
                min_ratio = []
                for idc, channel in enumerate(background.get("channels", [])):
                    current_signal = []
                    for sigch in signal:
                        if idc == int(sigch["path"].split("/")[2]):
                            current_signal = np.array(
                                sigch.get("value", {}).get("data", []), dtype=np.float32
                            )
                            break
                    if len(current_signal) == 0:
                        continue
                    current_bkg = []
                    for ch in channel["samples"]:
                        if len(current_bkg) == 0:
                            current_bkg = np.zeros(
                                shape=(len(ch["data"]),), dtype=np.float32
                            )
                        current_bkg += np.array(ch["data"], dtype=np.float32)
                    min_ratio.append(
                        np.min(
                            np.true_divide(
                                current_bkg[current_signal != 0.0],
                                current_signal[current_signal != 0.0],
                            )
                        )
                        if np.any(current_signal != 0.0)
                        else np.inf
                    )
                minimum_poi = (
                    -np.min(min_ratio).astype(np.float32)
                    if len(min_ratio) > 0
                    else -np.inf
                )

    if return_full_data:
        return signal, background, nb, delta_nb, workspace, model, data, minimum_poi

    return workspace, model, data


def objective_wrapper(
    data: np.ndarray, pdf: manager.pyhf.pdf, do_grad: bool
) -> Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]:
    """
    Prepare objective function and its gradient

    Args:
        data (``np.ndarray``): Input data
        pdf (``pyhf.pdf``): statistical model
        do_grad (``bool``): If ``True``, returns objective and gradient
          in a tuple.

    Returns:
        ``Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]``:
        Function to compute either only the objective function, ``do_grad=False``, or
        alongside with its gradient, ``do_grad=True``.

        .. warning::

            Gradient is only available for certain pyhf backends.
    """
    minimizer_kwargs, _ = manager.shim(
        manager.pyhf.infer.mle.twice_nll,
        manager.pyhf.tensorlib.astensor(data),
        pdf,
        pdf.config.suggested_init(),
        pdf.config.suggested_bounds(),
        None,
        do_grad,
        False,
    )

    def func(pars: np.ndarray) -> Union[np.array, Tuple[np.array, np.array]]:
        pars = manager.pyhf.tensorlib.astensor(pars)
        if do_grad:
            obj, grad = minimizer_kwargs["func"](pars)
            return np.array(obj), np.array(grad)

        return np.array(minimizer_kwargs["func"](pars))

    return func



"""pyhf plugin for spey interface"""

from typing import Optional, List, Text, Union, Callable, Tuple, Dict
import copy, warnings
import numpy as np

from spey.utils import ExpectationType
from spey.base.backend_base import BackendBase
from spey.base import ModelConfig
from .utils import objective_wrapper
from ._version import __version__
from . import manager
from .data import Base, SimpleModelData, FullStatisticalModelData

__all__ = ["UncorrelatedBackground", "FullStatisticalModel"]


def __dir__():
    return __all__


class ModelNotDefined(Exception):
    """Undefined model exception"""


class CombinationError(Exception):
    """Combination of the models are not possible"""


class PyhfInterface(BackendBase):
    """
    pyhf Interface. For details on input structure please see
    `this link <https://pyhf.readthedocs.io/en/v0.7.0/likelihood.html>`_

    Args:
        signal_yields (``Union[List[Dict], float, List[float]]``): Signal yields can be given in
          three different structure:

          * ``List[Dict]``: This is a ``JSONPATCH`` type of input.
          * ``float``: single float type of input will be used to create a single bin
            statistical model. This input expects both ``data`` and ``background_yields``
            inputs are ``float`` as well.
          * ``List[float]]``: This input type will be used to create an uncorrelated multi-bin
            statistical model. This input expects both ``data`` and ``background_yields`` inputs
            are ``List[float]]`` as well.

        data (``Union[Dict, float, List[float]]]``): Data input can be given in three different forms:

          * ``Dict``: This input is expected to be background only ``JSON`` based statistical
            model input. Please see the details from the link above.
          * ``float``: single float type of input will be used to create a single bin
            statistical model. This input expects both ``signal_yields`` and ``background_yields``
            inputs are ``float`` as well.
          * ``List[float]]``: This input type will be used to create an uncorrelated multi-bin
            statistical model. This input expects both ``signal_yields`` and ``background_yields``
            inputs are ``List[float]]`` as well.

        background_yields (``Union[float, List[float]]``, default ``None``): If ``data`` and
          ``signal_yields`` inputs are ``float`` or ``List[float]`` type, this input will be used
          to set the SM background yields in the statistical model. Not used when ``signal_yields``
          and ``data`` are in ``JSON`` format.
        absolute_background_unc (``Union[float, List[float]]``, default ``None``): If ``data`` and
          ``signal_yields`` inputs are ``float`` or ``List[float]`` type, this input will be used
          to set the absolute uncertainties in the SM background. Not used when ``signal_yields``
          and ``data`` are in ``JSON`` format.

    Example:

    .. code-block:: python3
        :linenos:

        >>> import spey

        >>> background_only = {
        ...     "channels": [
        ...         {
        ...             "name": "singlechannel",
        ...             "samples": [
        ...                 {
        ...                     "name": "background",
        ...                     "data": [50.0, 52.0],
        ...                     "modifiers": [
        ...                         {
        ...                             "name": "uncorr_bkguncrt",
        ...                             "type": "shapesys",
        ...                             "data": [3.0, 7.0],
        ...                         }
        ...                     ],
        ...                 }
        ...             ],
        ...         }
        ...     ],
        ...     "observations": [{"name": "singlechannel", "data": [51.0, 48.0]}],
        ...     "measurements": [{"name": "Measurement", "config": {"poi": "mu", "parameters": []}}],
        ...     "version": "1.0.0",
        ... }
        >>> signal = [
        ...     {
        ...         "op": "add",
        ...         "path": "/channels/0/samples/1",
        ...         "value": {
        ...             "name": "signal",
        ...             "data": [12.0, 11.0],
        ...             "modifiers": [{"name": "mu", "type": "normfactor", "data": None}],
        ...         },
        ...     }
        ... ]
        >>> statistical_model = spey.get_correlated_nbin_statistical_model(
        ...     analysis="simple_pyhf",
        ...     data=background_only,
        ...     signal_yields=signal,
        ... )
        >>> statistical_model.exclusion_confidence_level() # [0.9474850259721279]
    """

    name: Text = "pyhf.base"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = "0.1.0"
    """Spey version required for the backend"""
    doi: List[Text] = ["10.5281/zenodo.1169739", "10.21105/joss.02823"]
    """Citable DOI for the backend"""

    __slots__ = ["_model", "manager"]

    def __init__(self):
        self.manager = manager
        """pyhf Manager to handle the interface with pyhf"""

    @property
    def model(self) -> Base:
        """Retreive statistical model container"""
        if hasattr(self, "_model"):
            return self._model
        raise ModelNotDefined("Statistical model is not defined.")

    @property
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return self.model.isAlive

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: float = 10.0
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        return self.model.config(
            allow_negative_signal=allow_negative_signal, poi_upper_bound=poi_upper_bound
        )

    def expected_data(self, pars: List[float]) -> List[float]:
        r"""
        Compute the expected value of the statistical model

        Args:
            pars (``List[float]``): nuisance parameters, :math:`\theta` and
              parameter of interest, :math:`\mu`.

        Returns:
            ``List[float]``:
            Expected data of the statistical model
        """
        return self.model()[1].expected_data(pars)

    def get_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Generate function to compute :math:`\log\mathcal{L}(\mu, \theta)` where :math:`\mu` is the
        parameter of interest and :math:`\theta` are nuisance parameters.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and computes
            :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        # CHECK THE MODEL BOUNDS!!
        # POI Test needs to be adjusted according to the boundaries for sake of convergence
        # see issue https://github.com/scikit-hep/pyhf/issues/620#issuecomment-579235311
        # comment https://github.com/scikit-hep/pyhf/issues/620#issuecomment-579299831
        # NOTE During tests we observed that shifting poi with respect to bounds is not needed.
        _, model, data_org = self.model(expected=expected)

        return lambda pars: model.logpdf(
            pars, self.manager.pyhf.tensorlib.astensor(data_org if data is None else data)
        )[0]

    def get_hessian_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Currently Hessian of :math:`\log\mathcal{L}(\mu, \theta)` is only used to compute
        variance on :math:`\mu`. This method returns a callable function which takes fit
        parameters (:math:`\mu` and :math:`\theta`) and returns Hessian.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and
            returns Hessian of :math:`\log\mathcal{L}(\mu, \theta)`.
        """

        logpdf = self.get_logpdf_func(expected, data)

        if self.manager.backend == "jax":
            hess = self.manager.backend_accessor.hessian(logpdf)

            def func(pars: np.ndarray) -> np.ndarray:
                """Compute hessian of logpdf"""
                return np.array(hess(self.manager.pyhf.tensorlib.astensor(pars)))

            return func

        raise NotImplementedError(
            f"Hessian is not available in {self.manager.backend} backend"
        )

    def get_objective_function(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
        do_grad: bool = True,
    ) -> Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]:
        r"""
        Objective function is the function to perform the optimisation on. This function is
        expected to be twice negative log-likelihood, :math:`-2\log\mathcal{L}(\mu, \theta)`.
        Additionally, if available it canbe bundled with the gradient of twice negative log-likelihood.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit
            do_grad (``bool``, default ``True``): If ``True`` return objective and its gradient
              as ``tuple`` (subject to availablility) if ``False`` only returns objective function.

        Returns:
            ``Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]``:
            Function which takes fit parameters (:math:`\mu` and :math:`\theta`) and returns either
            objective or objective and its gradient.
        """

        if do_grad and not self.manager.grad_available:
            raise NotImplementedError(
                f"Gradient is not available for {self.manager.backend} backend."
            )

        _, model, data_org = self.model(expected=expected)

        return objective_wrapper(
            data=data_org if data is None else data, pdf=model, do_grad=do_grad
        )

    def get_sampler(self, pars: np.ndarray) -> Callable[[int], np.ndarray]:
        r"""
        Retreives the function to sample from.

        Args:
            pars (:obj:`np.ndarray`): fit parameters (:math:`\mu` and :math:`\theta`)

        Returns:
            ``Callable[[int], np.ndarray]``:
            Function that takes ``number_of_samples`` as input and draws as many samples
            from the statistical model.
        """
        _, model, _ = self.model()
        pdf = model.make_pdf(self.manager.pyhf.tensorlib.astensor(pars))

        def sampler(number_of_samples: int) -> np.ndarray:
            """
            Sample generator for the statistical model

            :param number_of_samples (`int`): number of samples to be drawn from the model
            :return `np.ndarray`: Sampled observations
            """
            return np.array(pdf.sample((number_of_samples,)))

        return sampler


class UncorrelatedBackground(PyhfInterface):
    """
    This backend initiates ``pyhf.simplemodels.uncorrelated_background``, forming an uncorrelated
    histogram structure with given inputs.

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[float]``): observations
        absolute_uncertainties (``List[float]``): absolute uncertainties on the background
    """

    name: Text = "pyhf.uncorrelated_background"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = "0.1.0"
    """Spey version required for the backend"""
    doi: List[Text] = PyhfInterface.doi
    """Citable DOI for the backend"""

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
        absolute_uncertainties: List[float],
    ):
        super().__init__()
        self._model = SimpleModelData(
            signal_yields, background_yields, data, absolute_uncertainties
        )


class FullStatisticalModel(PyhfInterface):
    """
    pyhf Interface. For details on input structure please see
    `this link <https://pyhf.readthedocs.io/en/v0.7.0/likelihood.html>`_

    Args:
        signal_patch (``List[Dict]``): Patch data for signal model. please see
            `this link <https://pyhf.readthedocs.io/en/v0.7.0/likelihood.html>`_ for details on
            the structure of the input.
        background_only_model (``Dict`` or ``Text``): This input expects background only data
            that describes the full statistical model for the background. It also accepts ``str``
            input which indicates the full path to the background only ``JSON`` file.

    Example:

    .. code-block:: python3
        :linenos:

        >>> import spey

        >>> background_only = {
        ...     "channels": [
        ...         {
        ...             "name": "singlechannel",
        ...             "samples": [
        ...                 {
        ...                     "name": "background",
        ...                     "data": [50.0, 52.0],
        ...                     "modifiers": [
        ...                         {
        ...                             "name": "uncorr_bkguncrt",
        ...                             "type": "shapesys",
        ...                             "data": [3.0, 7.0],
        ...                         }
        ...                     ],
        ...                 }
        ...             ],
        ...         }
        ...     ],
        ...     "observations": [{"name": "singlechannel", "data": [51.0, 48.0]}],
        ...     "measurements": [{"name": "Measurement", "config": {"poi": "mu", "parameters": []}}],
        ...     "version": "1.0.0",
        ... }
        >>> signal = [
        ...     {
        ...         "op": "add",
        ...         "path": "/channels/0/samples/1",
        ...         "value": {
        ...             "name": "signal",
        ...             "data": [12.0, 11.0],
        ...             "modifiers": [{"name": "mu", "type": "normfactor", "data": None}],
        ...         },
        ...     }
        ... ]
        >>> stat_wrapper = spey.get_backend("pyhf")
        >>> statistical_model = stat_wrapper(
        ...     analysis="simple_pyhf",
        ...     background_only_model=background_only,
        ...     signal_patch=signal,
        ... )
        >>> statistical_model.exclusion_confidence_level() # [0.9474850259721279]
    """

    name: Text = "pyhf"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = "0.1.0"
    """Spey version required for the backend"""
    doi: List[Text] = PyhfInterface.doi
    """Citable DOI for the backend"""

    def __init__(
        self,
        signal_patch: Dict,
        background_only_model: Union[Text, Dict],
    ):
        super().__init__()
        self._model = FullStatisticalModelData(signal_patch, background_only_model)

    def combine(self, other, **kwargs):
        """
        Combine full statistical models generated by pyhf interface

        Args:
            other (``FullStatisticalModel``): other statistical model to be combined with this model
            kwargs:

              pyhf specific inputs:

              * join (``str``, default ``None``): How to join the two workspaces.
                Pick from ``"none"``, ``"outer"``, ``"left outer"`` or "right outer".
              * merge_channels (``bool``): Whether or not to merge channels when performing the combine.
                This is only done with ``"outer"``, ``"left outer"``, and ``"right outer"`` options.

              non-pyhf specific inputs:

              * update_measurements (``bool``, default ``True``): In case the measurement name of two
                statistical models are the same, other statistical model's measurement name will be
                updated. If set to ``False`` measurements will remain as is.

              .. note::

                This model is ``"left"`` and other model is considered to be ``"right"``.

        Raises:
            ``CombinationError``: Raised if its not possible to combine statistical models.

        Returns:
            :obj:~spey-pyhf.interface.FullStatisticalModel:
            Combined statistical model.
        """
        assert isinstance(
            other, FullStatisticalModel
        ), f"Combination between {self.name} and {other.name} backends is not available."

        update_measurements = kwargs.pop("update_measurements", True)

        this_workspace = copy.deepcopy(self.model.workspace)
        other_workspace = copy.deepcopy(other.model.workspace)

        if update_measurements:
            warnings.warn(
                "Measurement names are identical which may create problems during combination."
                "The measurement name of the other statistical model will be updated. "
                "However, if this is not the desired action please set ``update_measurements`` to ``False``."
            )
            for this_measurement in this_workspace["measurements"]:
                for idy, other_measurement in enumerate(other_workspace["measurements"]):
                    if this_measurement["name"] == other_measurement["name"]:
                        other_workspace["measurements"][idy]["name"] += "_updated"

        try:
            combined_workspace = self.manager.pyhf.Workspace.combine(
                this_workspace, other_workspace, **kwargs
            )
        except Exception as err:
            raise CombinationError(
                "Unable to combine given background only models, they might not be compatible."
                "Please try with different `join` option : 'none', 'outer', 'left outer', or 'right outer'"
                " and `merge_channels` option."
            ) from err

        # Reorganise signal patch

        # Collect channel names for combined model
        combined_channel_names = [ch["name"] for ch in combined_workspace["channels"]]
        this_channel_names = [ch["name"] for ch in this_workspace["channels"]]
        other_channel_names = [ch["name"] for ch in other_workspace["channels"]]

        # genereta channel map for this signal patch to find the
        # location of the channel in the combined workspace
        this_channel_map = {
            this_channel_names[int(patch["path"].split("/")[2])]: patch
            for patch in copy.deepcopy(self.model.signal_patch)
        }
        # genereta channel map for other signal patch to find the
        # location of the channel in the combined workspace
        other_channel_map = {
            other_channel_names[int(patch["path"].split("/")[2])]: patch
            for patch in copy.deepcopy(other.model.signal_patch)
        }

        new_signal_patch = []
        to_remove = []
        for chid, chname in enumerate(combined_channel_names):
            current_map = None
            if chname in this_channel_map.keys():
                current_map = copy.deepcopy(this_channel_map)
            elif chname in other_channel_map.keys():
                current_map = copy.deepcopy(other_channel_map)
            else:
                raise CombinationError(
                    f"Signal patch is not weldefined. Can not find channel {chname} in either signal patches."
                )

            current_signal = current_map[chname]
            # update channel path
            current_path = current_signal["path"].split("/")[1:]
            current_path[1] = str(chid)
            current_signal["path"] = "/" + "/".join(current_path)

            if current_signal["op"] != "remove":
                # update signal name
                current_signal["value"][
                    "name"
                ] = f"{current_signal['value']['name']}_CID_{chid}"
                new_signal_patch.append(current_signal)
            else:
                to_remove.append(current_signal)

        # reorder the patches that will be removed i.e. lowest channel id should be at the end.
        to_remove.sort(key=lambda x: int(x["path"].split("/")[1:][1]), reverse=True)
        new_signal_patch += to_remove

        # Generate new statistical model
        return FullStatisticalModel(
            signal_patch=new_signal_patch, background_only_model=dict(combined_workspace)
        )



from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as f:
    long_description = f.read()

with open("src/spey_pyhf/_version.py", mode="r", encoding="UTF-8") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = ["pyhf==0.7.0", "spey>=0.1.0"]

docs = [
    "sphinx>=6.1.3",
    "sphinxcontrib-bibtex~=2.1",
    "sphinx-click",
    "sphinx_rtd_theme",
    "nbsphinx!=0.8.8",
    "sphinx-issues",
    "sphinx-copybutton>=0.3.2",
    "sphinx-togglebutton>=0.3.0",
    "myst-parser",
]

setup(
    name="spey-pyhf",
    version=version,
    description=("pyhf plugin for spey interface"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SpeysideHEP/spey-pyhf",
    project_urls={
        "Bug Tracker": "https://github.com/SpeysideHEP/spey-pyhf/issues",
    },
    download_url=f"https://github.com/SpeysideHEP/spey-pyhf/archive/refs/tags/v{version}.tar.gz",
    author="Jack Y. Araz",
    author_email=("jack.araz@durham.ac.uk"),
    license="MIT",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={
        "spey.backend.plugins": [
            "pyhf.uncorrelated_background = spey_pyhf.interface:UncorrelatedBackground",
            "pyhf = spey_pyhf.interface:FullStatisticalModel",
        ]
    },
    install_requires=requirements,
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    extras_require={
        "dev": ["pytest>=7.1.2", "pytest-cov>=3.0.0", "twine>=3.7.1", "wheel>=0.37.1"],
        "jax": ["jax>=0.3.35"],
        "doc": docs,
    },
)
